from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyfileconf.basemodels.config import ConfigBase

from pyfileconf.basemodels.file import ConfigFileBase
from pyfileconf.imports.models.statements.obj import ObjectImportStatement
from pyfileconf.assignments.models.statement import AssignmentStatement
from pyfileconf.io.file.write.asttosource import ast_node_to_source

class FunctionConfigFile(ConfigFileBase):
    """
    Represents config file on filesystem. Handles low-level functions for writing and reading config file
    """

    # lines to always import. pass import objects
    always_imports = [
        ObjectImportStatement.from_str('from pyfileconf import Selector', preferred_position='begin')
    ]

    # assignment lines to always include at beginning. pass assign objects
    always_assigns = [
        AssignmentStatement.from_str('s = Selector()', preferred_position='begin'),
    ]

    # always assign dict, where assigns will get added if item name matches dict key
    always_assign_with_names_dict = {
        'DataPipeline': [
            AssignmentStatement.from_str('cleanup_kwargs = dict(\n    \n)', preferred_position='end')
        ]
    }

    # class to use for interfacing with file
    # no need to override default
    # interface_class = ConfigFileInterface

    def save(self, config: 'ConfigBase'):
        # If empty cleanup kwargs on DataPipeline, then delete so can be replaced with dict() constructor
        # from always assigns
        if _should_replace_cleanup_kwargs(self, config):
            config.pop('cleanup_kwargs')

        super().save(config)


def _should_replace_cleanup_kwargs(config_file: FunctionConfigFile, config: 'ConfigBase') -> bool:
    return all([
        config_file.name == 'DataPipeline',
        'cleanup_kwargs' in config
    ]) and ast_node_to_source(config['cleanup_kwargs']) == 'None\n'


from typing import List, Tuple

from pyfileconf.io.file.load.lazy.base.impassign import ImportAssignmentLazyLoader
from pyfileconf.config.logic.load import _split_lines_into_import_and_assignment

class ConfigFileLoader(ImportAssignmentLazyLoader):

    def load(self) -> Tuple[dict, dict]:
        # Get ast, body, imports, assigns
        super().register()

        # Extract assignment portion of body
        self._file_body_to_assignment_body()

        # Parse assigns into config dict
        config_dict, annotation_dict = self.assigns.to_default_dict_and_annotation_dict()

        return config_dict, annotation_dict

    @property
    def assignment_body(self) -> List[str]:
        return self._try_getattr_else_call_func('_assignment_body', self.load)

    def _file_body_to_assignment_body(self):
        """
        Warning: Assumes that imports come first, then an assignment section. May have
        issues when imports come later
        Returns:

        """
        if self.body is not None:
            import_lines, assignment_lines = _split_lines_into_import_and_assignment(self.body)
            self._assignment_body = assignment_lines
        else:
            self._assignment_body = []




from typing import List, Optional
import ast

from pyfileconf.basemodels.container import Container
from mixins.repr import ReprMixin

from pyfileconf.exceptions.imports import NoImportMatchingNameException
from pyfileconf.imports.models.statements.obj import ObjectImportStatement
from pyfileconf.imports.models.statements.module import ModuleImportStatement
from pyfileconf.imports.models.statements.interfaces import (
    AnyImportStatementOrComment,
    Comment,
    ImportOrNone
)
from pyfileconf.imports.models.statements.rename import RenameStatementCollection
from pyfileconf.io.file.load.parsers.extname import extract_external_name_from_assign_value

class ImportStatementContainer(Container, ReprMixin):
    repr_cols = ['items']

    def __init__(self, items: List[AnyImportStatementOrComment]):
        self.items = items

    def __contains__(self, item: AnyImportStatementOrComment):
        for imp_or_comment in self:
            if item == imp_or_comment:
                return True

            if not isinstance(item, (ObjectImportStatement, ModuleImportStatement)) or \
               not isinstance(imp_or_comment, (ObjectImportStatement, ModuleImportStatement)):
                # only further handling is for actual imports, skip comments, etc.
                continue

            # handle where item is a subset of an import statement in the collection.

            # first handle where it's a module import statement
            # e.g. item is: import this
            # and import statement in collection is: import this, that
            if isinstance(item, ModuleImportStatement) and isinstance(imp_or_comment, ModuleImportStatement):
                matching_modules = all(mod in imp_or_comment.modules for mod in item.modules)
                matching_renames = all(rename in imp_or_comment.renames for rename in item.renames)
                if matching_modules and matching_renames:
                    return True

            # e.g. item is: from this import that
            # and import statement in collection is: from this import that, another
            if isinstance(item, ObjectImportStatement) and isinstance(imp_or_comment, ObjectImportStatement):
                matching_module = imp_or_comment.module == item.module
                matching_objs = all(mod in imp_or_comment.objs for mod in item.objs)
                matching_renames = all(rename in imp_or_comment.renames for rename in item.renames)
                if matching_module and matching_objs and matching_renames:
                    return True


        return False

    def __str__(self):
        return '\n'.join(str(imp) for imp in self)

    def obj_name_is_imported(self, obj_name: str) -> bool:

        obj_module: Optional[str]
        obj_name_parts = obj_name.split('.')
        if len(obj_name_parts) != 1:
            obj_module = '.'.join(obj_name_parts[:-1])
            obj_name = obj_name_parts[-1]
        else:
            # no . found, meaning object was imported by object import statement, no module name
            obj_module = None

        for imp_or_comment in self:
            if isinstance(imp_or_comment, Comment):
                continue # could not be imported from a comment
            if isinstance(imp_or_comment, ModuleImportStatement):
                if obj_module is None:
                    continue # object was imported by object import statement

                # object was imported by module import statement
                if (obj_module in imp_or_comment.modules) or (obj_module in imp_or_comment.renames.new_names):
                    return True # found matching module
            elif isinstance(imp_or_comment, ObjectImportStatement):
                if (obj_name in imp_or_comment.objs) or (obj_name in imp_or_comment.renames.new_names):
                    return True

        # failed all checks, obj not imported
        return False

    def get_import_for_ast_obj(self, obj_ast: ast.AST) -> ImportOrNone:
        possibly_imported_name = extract_external_name_from_assign_value(obj_ast)

        if possibly_imported_name is None:
            # Did not find any external names for this ast obj. Likely builtin.
            return None

        return self.get_import_for_module_or_obj_name(possibly_imported_name)

    def get_import_for_module_or_obj_name(self, name: str) -> ImportOrNone:
        found_import = False
        for imp in self:
            if isinstance(imp, ModuleImportStatement):
                if name in imp.modules:
                    # match on original module name
                    # set up for creating a new import
                    found_import = True
                    renames = None
                    modules = [name]
                elif name in imp.renames.new_names:
                    # match on renamed module name
                    # set up for creating a new import
                    found_import = True
                    renames = RenameStatementCollection(
                        # Pull the rename matching this name
                        [rename for rename in imp.renames if rename.new_name == name]
                    )
                    # grab original module matching this rename
                    modules = [renames.reverse_name_map[name]]
                if found_import:
                    # May be multiple modules imported in this one statement. Create a new statement with just this module
                    return ModuleImportStatement(modules=modules, renames=renames)
            elif isinstance(imp, ObjectImportStatement):
                if name in imp.objs:
                    # match on original object name
                    # set up for creating a new import
                    found_import = True
                    renames = None
                    objs = [name]
                elif name in imp.renames.new_names:
                    # match on renamed object name
                    # set up for creating a new import
                    found_import = True
                    renames = RenameStatementCollection(
                        # Pull the rename matching this name
                        [rename for rename in imp.renames if rename.new_name == name]
                    )
                    # grab original object matching this rename
                    objs = [renames.reverse_name_map[name]]
                if found_import:
                    # May be multiple objects imported in this one statement. Create a new statement with just this object
                    return ObjectImportStatement(objs, module=imp.module, renames=renames)
        raise NoImportMatchingNameException(f'could not find name {name} in {self}')

    @property
    def imported_names(self) -> List[str]:
        imported_names: List[str] = []
        for imp_or_comment in self:
            if isinstance(imp_or_comment, Comment):
                continue # could not be imported from a comment
            if isinstance(imp_or_comment, ModuleImportStatement):
                imported_names += imp_or_comment.modules
                imported_names += imp_or_comment.renames.new_names
            elif isinstance(imp_or_comment, ObjectImportStatement):
                imported_names += imp_or_comment.objs
                imported_names += imp_or_comment.renames.new_names
                imported_names.append(imp_or_comment.module)
        return list(set(imported_names)) # remove duplicates

    @property
    def modules(self):
        modules = []
        for item in self:
            # object import statements have one module
            if isinstance(item, ObjectImportStatement):
                modules.append(item.module)
            # module import statements can have multiple modules
            elif isinstance(item, ModuleImportStatement):
                modules += item.modules

        return modules

    @property
    def objs(self):
        objs = []
        for item in self:
            # only object import statements have objects
            if isinstance(item, ObjectImportStatement):
                objs += item.objs

        return objs

    # @property
    # def renames(self):
    #     renames = {}
    #     for item in self:
    #         item: AnyImportStatement
    #         renames[] = item.renames
    #
    #     return renames




from typing import Union

from pyfileconf.pipelines.models.collection import PipelineCollection
from pyfileconf.basemodels.registrar import Registrar
from pyfileconf.views.object import ObjectView

ObjectViewOrCollection = Union[ObjectView, PipelineCollection]

class PipelineRegistrar(Registrar):
    collection_class = PipelineCollection

    def get(self, section_path_str: str):
        obj_view_or_collection: ObjectViewOrCollection = super().get(section_path_str)

        # Nothing further needed for collections
        if isinstance(obj_view_or_collection, PipelineCollection):
            return obj_view_or_collection

        # Need to load ObjectView, return actual object
        return obj_view_or_collection.load()








from typing import Callable, Optional, Type, Sequence

from pyfileconf.basemodels.config import ConfigBase
from pyfileconf.imports.logic.load.func import function_args_as_dict
from pyfileconf.config.models.file import FunctionConfigFile

class FunctionConfig(ConfigBase):

    config_file_class = FunctionConfigFile

    @classmethod
    def from_file(cls, filepath: str, name: str = None,
                  klass: Optional[Type] = None, always_import_strs: Optional[Sequence[str]] = None,
                  always_assign_strs: Optional[Sequence[str]] = None):
        file = cls.config_file_class(
            filepath,
            name=name,
            klass=klass,
            always_import_strs=always_import_strs,
            always_assign_strs=always_assign_strs
        )
        return file.load(cls)

    def for_function(self, func: Callable) -> dict:
        """
        Strips out items of config which are not applicable to function. Returns dictionary
        of config items for passing to the function.

        Args:
            func: func for which to filter out config items

        Returns: dict, applicable config for func
        """
        # Only pass items in config which are arguments of function
        func_kwargs = function_args_as_dict(func)
        return {key: value for key, value in self.items() if key in func_kwargs}


import ast
from typing import Optional, Union

import black

from pyfileconf.sectionpath.sectionpath import SectionPath


def add_item_into_nested_dict_at_section_path(nested_dict: dict, section_path: SectionPath, item: str,
                                              add_as_ast_name: bool = True):
    """
    :Notes:
        Inplace

    :param nested_dict:
    :param section_path:
    :param item:
    :return:
    """
    current_dict = nested_dict
    for i, section in enumerate(section_path):
        if i <= len(section_path) - 2:
            # Up until the next to last item, pull out the nested dicts, creating if necessary
            try:
                current_dict = current_dict[section]
                # Section already exists, keep navigating deeper
            except KeyError:
                # Section does not yet exist, create it, then navigate into it
                current_dict[section] = {}
                current_dict = current_dict[section]
        if i == len(section_path) - 1:
            # Reached final section. Because we only went up until the next to last item pulling out
            # nested dicts, current_dict should now be the dict which contains the container for
            # the item to be stored.

            # Need to create entry in this section.
            new_entry: Union[str, ast.Name]
            if add_as_ast_name:
                new_entry = ast.Name(item)
            else:
                new_entry = item

            # First see if there is already an existing section
            try:
                last_container = current_dict[section]
            except KeyError:
                # No last section, create it
                current_dict[section] = []
                last_container = current_dict[section]

            if not isinstance(last_container, list):
                raise ValueError('trying to insert item into existing dictionary section, must be existing '
                                 'list section or non-existent section')

            last_container.append(new_entry)


def create_dict_assignment_str_from_nested_dict_with_ast_names(nested_dict: dict, existing_str: Optional[str] = None,
                                                               assign_to_name: str = 'pipeline_dict') -> str:
    if existing_str is None:
        use_str = f'{assign_to_name} = {{'
    else:
        use_str = ''

    for key, value in nested_dict.items():
        use_str += f'"{key}": '
        if isinstance(value, ast.Name):
            use_str += value.id + ','
        elif isinstance(value, str):
            use_str += value + ','
        elif isinstance(value, list):
            use_str += '['
            for item in value:
                use_str += item.id + ','
            use_str += '],'
        elif isinstance(value, dict):
            use_str += '{'
            use_str += create_dict_assignment_str_from_nested_dict_with_ast_names(value, use_str)
            use_str += '},'
        else:
            raise ValueError(f'could not parse object {value} of type {type(value)} in '
                             f'creating pipeline dict str')

    if existing_str is None:
        # Must be finishing top-level operation, finish str
        use_str += '}'
        # Pretty print the string using black formatter
        use_str = pretty_format_str(use_str)

    return use_str


def pretty_format_str(string: str) -> str:
    fm = black.FileMode()
    out_str = black.format_str(string, mode=fm)
    return out_str


from typing import Any


def combine_items_into_list_whether_they_are_lists_or_not_then_extract_from_list_if_only_one_item(
    item1: Any, item2: Any
) -> Any:
    items = combine_items_into_list_whether_they_are_lists_or_not(item1, item2)
    if len(items) == 1:
        return items[0]

    return items


def combine_items_into_list_whether_they_are_lists_or_not(
    item1: Any, item2: Any
) -> list:
    item1_list = _to_list(item1)
    item2_list = _to_list(item2)
    return [*item1_list, *item2_list]


def _to_list(item: Any) -> list:
    if item is None:
        return []
    if isinstance(item, list):
        return item
    elif isinstance(item, tuple):
        return list(item)
    else:
        return [item]



from typing import Tuple, List, Union, Sequence
from copy import deepcopy

ListOfStrs = List[str]
TwoNoneTuple = Tuple[None, None]
TwoStrTuple = Tuple[str, str]
TwoTupleNoneOrStr = Union[TwoStrTuple, TwoNoneTuple]

def _split_lines_into_import_and_assignment(lines: ListOfStrs, strip_lines=True) -> Tuple[ListOfStrs, ListOfStrs]:
    # TODO [#2]: deal with later imports
    #
    # Imports can currently only be read from the top of the file

    # TODO [#3]: deal with use of import other than the word import

    import_section = []
    assignment_section = []
    # We can have three cases for the following two booleans:
    # The file starts in the import section
    # Then, we enter the whitespace section between import and assignments. Still consider this the import section.
    # Finally, once text is detected again after import section,
    # assignment section starts (in_import_section=False)
    in_import_section = True
    for i, line in enumerate(lines):

        if strip_lines:
            line = line.strip('\n')

        # Handle import section and add whitespace below import section
        if in_import_section:
            if 'import' in line or _is_whitespace_line(line):
                import_section.append(line)
            else:
                in_import_section = False  # no longer dealing with imports, must have assignment line

        # Assignment section, just output
        if not in_import_section:
            assignment_section.append(line)

    # Import section was handled as both import section and following whitespace. Strip following whitespace.
    import_section = _strip_whitespace_lines_from_end_of_lines(import_section)

    return import_section, assignment_section

def _split_assignment_line_into_variable_name_and_assignment(line: str) -> TwoTupleNoneOrStr:

    # handle whitespace lines
    if '=' not in line:
        return None, None

    # Split into two parts by the first equals (handle multiple equals). Treating further equals as string
    # TODO [#4]: handle multiple assignments such as this = that = 5
    split_line = line.split('=')
    parts: Sequence[str]
    if len(split_line) > 2:
        parts = split_line[0], '='.join(split_line[1:])
    else:
        parts = split_line

    variable, value = tuple(item.strip() for item in parts)
    return variable, value

def _is_whitespace_line(line: str) -> bool:
    return line.strip() == ''

def _strip_whitespace_lines_from_end_of_lines(lines: List[str]) -> List[str]:

    # Handle empty list
    if lines == []:
        return []

    reversed_lines = deepcopy(lines)
    reversed_lines.reverse()

    # starting from end first. Discard all lines until non-whitespace is found
    for line_index, line in enumerate(reversed_lines):
        if not _is_whitespace_line(line):
            break

    output_lines = reversed_lines[line_index:]
    output_lines.reverse() # back to first first

    return output_lines




import sys
from importlib._bootstrap import _find_spec
from importlib.util import resolve_name
from importlib.machinery import ModuleSpec


def get_filepath_from_module_str(module_str: str, location_section_path_str: str=None):
    spec: ModuleSpec = find_spec(module_str, package=location_section_path_str)
    if not spec.has_location:
        raise ValueError(f'could not find location of module {module_str}, package {location_section_path_str}')

    return spec.origin

def find_spec(name, package=None) -> ModuleSpec:
    """Return the spec for the specified module.

    Note: this is a modification of importlib.util.find_spec which does not
    import parent packages. It instead recursively searches through the parents.

    First, sys.modules is checked to see if the module was already imported. If
    so, then sys.modules[name].__spec__ is returned. If that happens to be
    set to None, then ValueError is raised. If the module is not in
    sys.modules, then sys.meta_path is searched for a suitable spec with the
    value of 'path' given to the finders. None is returned if no spec could
    be found.

    If the name is for submodule (contains a dot), find_spec is called on the
    parent module to determine the submodule_search_locations.

    The name and package arguments work the same as importlib.import_module().
    In other words, relative module names (with leading dots) work.

    """
    fullname = resolve_name(name, package) if name.startswith('.') else name
    if fullname not in sys.modules:
        parent_name = fullname.rpartition('.')[0]
        if parent_name:
            # Use builtins.__import__() in case someone replaced it.
            parent = find_spec(parent_name, package=package)
            return _find_spec(fullname, parent.submodule_search_locations)
        else:
            return _find_spec(fullname, None)
    else:
        module = sys.modules[fullname]
        if module is None:
            return None
        try:
            spec = module.__spec__
        except AttributeError:
            raise ValueError('{}.__spec__ is not set'.format(name)) from None
        else:
            if spec is None:
                raise ValueError('{}.__spec__ is None'.format(name))
            return spec


"""
Contains the hooks which may be attached to in creating plugins
"""
from typing import Any, Dict, Sequence, List, Tuple, Optional, TYPE_CHECKING

from pyfileconf.runner.models.runner import Runner

if TYPE_CHECKING:
    from pyfileconf.main import PipelineManager

import pluggy

from pyfileconf.runner.models.interfaces import RunnerArgs, ResultOrResults

hookspec = pluggy.HookspecMarker("pyfileconf")


@hookspec
def pyfileconf_iter_get_cases(
    config_updates: Sequence[Dict[str, Any]]
) -> List[Tuple[Dict[str, Any], ...]]:
    """
    Called in PipelineManager.run_iter and IterativeRunner to take the user passed
    config updates and return the config cases to be run

    :param config_updates: list of kwarg dictionaries which would normally be provided to .update
    :return: config cases to be run
    """


@hookspec
def pyfileconf_iter_modify_cases(cases: List[Tuple[Dict[str, Any], ...]]):
    """
    Called in PipelineManager.run_iter and IterativeRunner to take the collected
    config cases to be run and modify them in place.

    :param cases: list of tuples of kwarg dictionaries which would normally be provided to .update
    :return: None
    """


@hookspec
def pyfileconf_pre_run(
    section_path_str_or_list: RunnerArgs, pm: "PipelineManager"
) -> Optional[RunnerArgs]:
    """
    Called at the beginning of PipelineManager.run. Can optionally return
    additional section paths to run. If section_path_str_or_list is a
    list then it can also be modified in place.

    :param section_path_str_or_list: section paths which were passed to
        PipelineManager.run
    :return: additional sections/functions to run, if any
    """


@hookspec
def pyfileconf_post_run(
    results: ResultOrResults, runner: Runner
) -> Optional[ResultOrResults]:
    """
    Called at the end of PipelineManager.run. Can optionally return
    additional results which will be appended to the results list. If results is mutable
    then it can also be modified in place.

    :param results: results from running section/function
    :return: additional results, if any
    """


@hookspec
def pyfileconf_pre_update(
    pm: "PipelineManager",
    d: dict = None,
    section_path_str: str = None,
    kwargs: Dict[str, Any] = None,
) -> Optional[Dict[str, Any]]:
    """
    Called at the beginning of PipelineManager.update. Can optionally return
    a dictionary of updates which will be used to update the passed dictionary.
    Can also modify the passed dictionaries in place.

    :param d: dictionary of config updates
    :param section_path_str: section path of config to be updated
    :param kwargs: dictionary of config updates
    :return: optional updates to config updates
    """


@hookspec
def pyfileconf_post_update(
    pm: "PipelineManager",
    d: dict = None,
    section_path_str: str = None,
    kwargs: Dict[str, Any] = None,
):
    """
    Called at the end of PipelineManager.update.

    :param d: dictionary of config updates
    :param section_path_str: section path of config which was updated
    :param kwargs: dictionary of config updates
    :return: None
    """



import builtins
from typing import Any
import inspect
from types import FrameType
from typing import cast


def get_caller_frame() -> FrameType:
    """Return the calling function's name."""
    return cast(FrameType, cast(FrameType, inspect.currentframe()).f_back)


def get_caller_filepath(caller_levels: int = 2) -> str:
    this_frame = get_caller_frame()
    frame = this_frame
    for _ in range(caller_levels):
        frame = frame.f_back
    return frame.f_code.co_filename


def _is_builtin(value: Any) -> bool:
    if value is None:
        return True  # None won't return True from the following check

    builtin_types = [
        getattr(builtins, d)
        for d in dir(builtins)
        if isinstance(getattr(builtins, d), type)
    ]
    return type(value) in builtin_types


def _is_str_matching_builtin_type(str_value: str) -> bool:
    """
    should pass 'str', 'int', etc.
    """
    builtin_types = [
        getattr(builtins, d)
        for d in dir(builtins)
        if isinstance(getattr(builtins, d), type)
    ]
    return f"<class '{str_value}'>" in [str(bt) for bt in builtin_types]



import inspect
from functools import partial
from typing import cast, List, Type, Dict, Tuple

from pyfileconf.data.models.collection import SpecificClassCollection
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf.selector.logic.exc.typo import (
    handle_pipeline_manager_not_loaded_or_typo,
    handle_known_typo_at_end_of_section_path_str,
    handle_known_typo_after_pipeline_manager_name
)
from pyfileconf.selector.models.selector import Selector
from pyfileconf.exceptions.pipelinemanager import PipelineManagerNotLoadedException
from copy import deepcopy


class ItemView:
    """
    Class for representing a pipeline manager unit (function, pipeline, data source) without needing
    that unit to be loaded into pipeline manager. Allows using selector in config files and in app,
    by delaying looking up the item until an attribute/method is accessed or item is called.
    """

    def __init__(self, section_path_str: str, selector: Selector):
        self.section_path_str = section_path_str
        self._section_path_str = section_path_str  # for compatibility with real items which have this attribute added
        self.selector = selector
        self._is_item_view = True

    def __getattr__(self, item):
        from pyfileconf.main import PipelineManager
        full_section_path_str = self.section_path_str + '.' + item
        if full_section_path_str in self.selector:
            # This is an item, return an item view for it
            return ItemView(full_section_path_str, self.selector)

        # Not an item.
        # Must be either an attribute of an item, or a typo. Actually look up the item now

        try:
            actual_item = self.item
        except PipelineManagerNotLoadedException:
            # Dealing with typos is difficult because if this is a typo and we are reaching here,
            # if PipelineManager.load() has not been run yet, we can't know the attributes of the items,

            # check whether this is an item attribute or a typo, and raise error if needed
            return handle_pipeline_manager_not_loaded_or_typo(full_section_path_str, self.selector._managers)

        # TODO [#18]: refactor item view lookup errors
        #
        #  So that there is a better way of catching a known typo than python throwing a RecursionError

        except RecursionError:
            # We are landing here when there is a typo in after the pipeline manager selection of a longer
            # section path, e.g if "this" was portfolio manager name: s.this.tpyo.would.cause.this
            return handle_known_typo_after_pipeline_manager_name(full_section_path_str)

        try:
            result = getattr(actual_item, item)
        except (KeyError, AttributeError) as e:
            ### TEMP - having issues with _is_selector
            raise e
            ### END TEMP
            return handle_known_typo_at_end_of_section_path_str(full_section_path_str)

        # Got attribute of actual item
        # If this happened while running another item, add to dependencies
        self._add_to_config_dependencies_if_necessary()

        return result

    def __setattr__(self, key: str, value):
        # Set these items on ItemView itself
        item_view_set_attrs = [
            'section_path_str',
            'selector',
            '_is_item_view',
            '_section_path_str',
        ]
        if key in item_view_set_attrs:
            super().__setattr__(key, value)
        else:
            # Set any others on the original object via updating PipelineManager config
            self.selector._set_attr_for_item(self.section_path_str, key, value)

    def __dir__(self):
        exposed_properties = [
            'type',
            'item'
        ]
        collection_attrs = self.selector._get_dir_for_section_path(self.section_path_str)
        return exposed_properties + collection_attrs

    def __call__(self, *args, **kwargs):
        # When calling, assume user always wants the real item
        actual_item = self.selector._get_real_item(self.section_path_str)
        # If this happened while running another item, add to dependencies
        self._add_to_config_dependencies_if_necessary()
        if isinstance(actual_item, partial):
            # Got something in the general registrar, function or class
            func = actual_item
        elif isinstance(actual_item, self._specific_classes):
            # Got specific registrar class
            # Need to look up the execute attribute and apply section path str
            actual_item._section_path_str = self.section_path_str
            collection = self._specific_class_collection_map[type(actual_item)]
            execute_attr = collection.execute_attr
            func = getattr(actual_item, execute_attr)
        else:
            cannot_parse_error = ValueError(f'could not parse actual item, expected partial, '
                                            f'specific class, or method of specific class. '
                                            f'Got {actual_item} of type {type(actual_item)}')
            try:
                orig_item = actual_item.__self__
                # Is bound method
            except AttributeError:
                # Is not bound method
                raise cannot_parse_error

            if not isinstance(orig_item, self._specific_classes):
                # Is bound method, but not for one of defined specific classes
                raise cannot_parse_error

            # Got specific class method
            # Add section path to original item and then set method to be called
            orig_item_sp = SectionPath.from_section_str_list(SectionPath(self.section_path_str)[:-1])
            orig_item_sp_str = orig_item_sp.path_str
            orig_item._section_path_str = orig_item_sp_str
            func = actual_item

        result = func(*args, **kwargs)
        if isinstance(actual_item, partial) and inspect.isclass(actual_item.func):
            # Got a class in the general collection, running these generates
            # an instance of the object so add the section path in this case
            result._section_path_str = self.section_path_str
        return result

    def __deepcopy__(self, memodict={}):
        return deepcopy(self.item, memodict)

    def __eq__(self, other):
        equal_attrs = ['section_path_str']

        for equal_attr in equal_attrs:
            if not hasattr(other, equal_attr): # other object doesn't have this property, must not be equal
                return False
            if getattr(self, equal_attr) != getattr(other, equal_attr):
                return False

        # passed all checks, must be equal
        return True

    def __repr__(self):
        repr_cols = ['section_path_str']

        repr_col_strs = [f'{col}={getattr(self, col, None).__repr__()}' for col in repr_cols]
        repr_col_str = f'({", ".join(repr_col_strs)})'

        return f'<{type(self).__name__}{repr_col_str}>'

    @property
    def type(self):
        return self.selector.get_type(self.section_path_str)

    @property
    def item(self):
        return self.selector._get_real_item(self.section_path_str)

    def _add_to_config_dependencies_if_necessary(self):
        from pyfileconf.context import context
        context.add_config_dependency_for_currently_running_item_if_exists(self.section_path_str, force_update=True)

    @property
    def _specific_class_collection_map(self) -> Dict[Type, SpecificClassCollection]:
        class_collection_map: Dict[Type, SpecificClassCollection] = {}
        for manager_name, manager_dict in self.selector._structure.items():
            for collection_name, collection in manager_dict.items():
                if collection_name == '_general':
                    continue
                class_collection_map[collection.klass] = collection
        return class_collection_map

    @property
    def _specific_classes(self) -> Tuple[Type, ...]:
        return tuple(self._specific_class_collection_map.keys())


def _is_item_view(obj) -> bool:
    is_item_view = getattr(obj, '_is_item_view', False)
    if is_item_view:
        obj = cast(ItemView, obj)
    return is_item_view



from pyfileconf.io.file.load.active.userdef import get_user_defined_dict_from_filepath
from pyfileconf.io.file.load.lazy.base.impassign import ImportAssignmentLazyLoader

class ActiveConfigFileLoader(ImportAssignmentLazyLoader):

    def load(self):
        from pyfileconf import context
        # Get ast, imports, assigns
        super().register()

        # Actually import module
        context.file_is_currently_being_loaded = True
        self._user_defined_dict = get_user_defined_dict_from_filepath(self.filepath)
        context.file_is_currently_being_loaded = False

        return self._user_defined_dict

    @property
    def user_defined_dict(self) -> dict:
        return self._try_getattr_else_call_func('_user_defined_dict', self.load)


import ast

from mixins.propertycache import SimplePropertyCacheMixin
from pyfileconf.imports.models.statements.interfaces import (
    AnyImportStatement,
    ObjectImportStatement,
    ModuleImportStatement
)
from pyfileconf.imports.models.statements.container import ImportStatementContainer
from pyfileconf.io.file.load.parsers.extname import extract_external_name_from_assign_value
from pyfileconf.io.file.load.parsers.kwargs import extract_keywords_from_ast_by_name
from pyfileconf.io.func.load.config import FunctionConfigExtractor
from mixins.repr import ReprMixin

class ObjectView(SimplePropertyCacheMixin, ReprMixin):
    repr_cols = ['name', 'obj_ast', 'section_path_str']

    def __init__(self, obj_ast: ast.AST, import_statement: AnyImportStatement=None,
                 section_path_str: str=None):
        self.obj_ast = obj_ast
        self.import_statement = import_statement
        self.section_path_str = section_path_str

    def load(self):
        # executes import
        if self.import_statement is not None:
            self._item = _execute_import_get_obj_from_result(self.import_statement, self.name)

        return self._item

    @property
    def default_config(self):
        return self._try_getattr_else_call_func('_default_config', self._get_default_config)

    @property
    def item(self):
        return self._try_getattr_else_call_func('_item', self.load)

    @property
    def module(self) -> str:
        return self._try_getattr_else_call_func('_module', self._set_module)

    @property
    def name(self) -> str:
        return self._try_getattr_else_call_func('_name', self._set_name)

    @property
    def output_name(self) -> str:
        try:
            return self._output_name
        except AttributeError:
            return self.name

    @classmethod
    def from_ast_and_imports(cls, obj_ast: ast.AST, import_statements: ImportStatementContainer,
                             section_path_str: str=None):
        import_statement = import_statements.get_import_for_ast_obj(obj_ast)

        return cls(
            obj_ast,
            import_statement=import_statement,
            section_path_str=section_path_str
        )

    def _set_module(self):
        if self.import_statement is None:
            self._module = None

        if isinstance(self.import_statement, ObjectImportStatement):
            self._module = self.import_statement.module
        elif isinstance(self.import_statement, ModuleImportStatement):
            # have already ensured in creating ObjectView that there should be only one module
            # in the module import statement. So just take the first module
            self._module = self.import_statement.modules[0]
        else:
            raise ValueError(f'expected import statement to be ObjectImportStatement or ModuleImportStatement.'
                             f' Got {self.import_statement} of type {type(self.import_statement)}')

    def _set_name(self):
        name = extract_external_name_from_assign_value(self.obj_ast)
        if name == 'DataPipeline':
            # DataPipeline may be passed in pipeline dict. Then need to get name from DataPipeline constructor
            ast_name = extract_keywords_from_ast_by_name(self.obj_ast, 'name')['name']
            self._output_name = ast_name.s
        self._name = name

    def _get_default_config(self):
        self._default_config = FunctionConfigExtractor(self).extract_config(self.name)


def _execute_import_get_obj_from_result(import_statement: AnyImportStatement, obj_name: str):
    # Should be just a single module or object, can take first of either.
    result = import_statement.execute()[0]

    if isinstance(import_statement, ObjectImportStatement):
        # Result is the object itself
        return result
    elif isinstance(import_statement, ModuleImportStatement):
        # Result is the module. Get the object from the module
        return getattr(result, obj_name)
    else:
        raise ValueError(f'expected import statement to be ObjectImportStatement or ModuleImportStatement.'
                         f' Got {import_statement} of type {type(import_statement)}')


from pyfileconf.logic.userinput import _is_begin_str

class OrderPreferenceMixin:

    @property
    def preferred_position(self):
        return 'begin' if self.prefer_beginning else 'end'

    @preferred_position.setter
    def preferred_position(self, position_str: str):
        self.prefer_beginning = self._is_begin_str(position_str)

    def _is_begin_str(self, begin_str: str) -> bool:
        return _is_begin_str(begin_str)


import ast
from typing import Union

from pyfileconf.exceptions.imports import ExtractedIncorrectTypeOfImportException
from pyfileconf.imports.models.statements.interfaces import (
    ObjectImportStatement,
    ModuleImportStatement,
    AnyImportStatement
)
from pyfileconf.imports.models.statements.container import ImportStatementContainer

AnyImportStatementOrNone = Union[AnyImportStatement, None]
ObjectImportStatementOrNone = Union[ObjectImportStatement, None]
ModuleImportStatementOrNone = Union[ModuleImportStatement, None]

class ImportExtractor(ast.NodeVisitor):

    def __init__(self):
        self.imports = ImportStatementContainer([])

    def visit_Import(self, node):
        self.imports.append(
            ModuleImportStatement.from_ast_import(node)
        )

    def visit_ImportFrom(self, node):
        self.imports.append(
            ObjectImportStatement.from_ast_import_from(node)
        )

def extract_imports_from_ast(module: ast.Module) -> ImportStatementContainer:
    ie = ImportExtractor()
    ie.visit(module)
    return ie.imports

def extract_obj_import_from_ast(module: ast.Module) -> ObjectImportStatementOrNone:
    import_container = extract_imports_from_ast(module)
    imp = _extract_single_import_from_container(import_container)
    _validate_import_type(imp, ObjectImportStatement)

    return imp

def extract_module_import_from_ast(module: ast.Module) -> ModuleImportStatementOrNone:
    import_container = extract_imports_from_ast(module)
    imp = _extract_single_import_from_container(import_container)
    _validate_import_type(imp, ModuleImportStatement)

    return imp

def _extract_single_import_from_container(import_container: ImportStatementContainer) -> AnyImportStatementOrNone:
    if len(import_container) == 0:
        return None

    if len(import_container) > 1:
        raise ValueError(f'expected to extract one import, got {len(import_container)} imports: '
                         f'{import_container.items}')

    return import_container[0]

def _validate_import_type(import_statement: AnyImportStatementOrNone, enforce_type: type=ObjectImportStatement):
    if import_statement is None:
        return

    if not isinstance(import_statement, enforce_type):
        raise ExtractedIncorrectTypeOfImportException(
            f'expected import of type {enforce_type}, got {import_statement} of '
            f'type {type(import_statement)}'
        )


from copy import deepcopy

from pyfileconf import Selector, PipelineManager
from pyfileconf.iterate import IterativeRunner
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf import context
from tests.input_files.amodule import SecondExampleClass, a_function
from tests.input_files.mypackage.cmodule import ExampleClass
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST, SAME_CLASS_CONFIG_DICT_LIST, \
    DIFFERENT_CLASS_CONFIG_DICT_LIST


class TestUpdateContext(PipelineManagerTestBase):

    def test_update_currently_running_section_path_str_to_get_dependencies_while_running(self):
        self.write_example_class_dict_to_file()
        ccdl = deepcopy(CLASS_CONFIG_DICT_LIST)
        for config_dict in ccdl:
            config_dict['execute_attr'] = 'dependent_call_with_context_update'
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=ccdl
        )
        pipeline_manager.load()
        pipeline_manager.create('example_class.stuff.data2')
        pipeline_manager.create('example_class.stuff.data3')
        pipeline_manager.create('example_class.stuff.data4')
        pipeline_manager.create('ec', ExampleClass)
        pipeline_manager.create('ec2', ExampleClass)
        pipeline_manager.create('ec3', ExampleClass)
        pipeline_manager.create('af', a_function)
        pipeline_manager.create('af2', a_function)
        pipeline_manager.create('af3', a_function)
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        result = pipeline_manager.run(iv)
        assert context.config_dependencies == context.force_update_dependencies \
               == self.expect_pm_1_specific_class_4_depends_on_pm_1_specific_class_3_class_1_2_function_1_2
        assert context.currently_running_section_path_str is None





from copy import deepcopy
from typing import Any, List


class Container:
    items: List[Any] = []

    def __iter__(self):
        for item in self.items:
            yield item

    def __getitem__(self, item):
        return self.items[item]

    def __contains__(self, item):
        return item in self.items

    def __len__(self):
        return len(self.items)

    def __add__(self, other):
        cls = type(self)
        if not isinstance(other, cls):
            raise TypeError(f'must be {cls}, not {type(other)}')

        return cls(self.items + other.items)

    def append(self, item):
        self.items.append(item)

    def extend(self, items):
        self.items.extend(items)

    def insert(self, index, item):
        self.items.insert(index, item)

    def copy(self):
        return deepcopy(self)

    def add_if_missing(self, item):
        """
        Checks if item is in container. If so, doesn't do anything.

        Assuming item is not in container,
        checks for prefer_beginning attribute of item. If True, insert at position 0.
        If False, or doesn't have attr, append to end.

        """
        add_to_begin = getattr(item, 'prefer_beginning', False)

        if add_to_begin:
            self.insert_if_missing(0, item)
        else:
            self.append_if_missing(item)

    def append_if_missing(self, item):
        if item not in self:
            self.append(item)

    def insert_if_missing(self, index, item):
        if item not in self:
            self.insert(index, item)




from typing import Callable, Union, List, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from pyfileconf.pipelines.models.collection import PipelineCollection

from pyfileconf.basemodels.pipeline import Pipeline
from pyfileconf.views.object import ObjectView

PipelineOrFunction = Union[Pipeline, Callable]
PipelineOrFunctionOrCollection = Union[PipelineOrFunction, 'PipelineCollection']
PipelinesOrFunctions = List[PipelineOrFunction]
PipelinesOrFunctionsOrCollections = List[PipelineOrFunctionOrCollection]

# TODO [#27]: remove type ignores for recursive type definitions once mypy supports them
#
# See https://github.com/python/mypy/issues/731
PipelineDict = Dict[str, Union[PipelinesOrFunctions, 'PipelineDict']]  # type: ignore
PipelineDictOrPipelineOrFunction = Union[PipelineDict, PipelineOrFunction]  # type: ignore
PipelineDictsOrPipelinesOrFunctions = List[PipelineDictOrPipelineOrFunction]  # type: ignore

StrList = List[str]
ObjectViewOrCollection = Union[ObjectView, 'PipelineCollection']


import ast
from typing import TYPE_CHECKING, Tuple, Optional, cast

from pyfileconf.exceptions.imports import NoImportStatementException

if TYPE_CHECKING:
    from pyfileconf.views.object import ObjectView
from importlib.util import resolve_name

from pyfileconf.imports.models.statements.interfaces import (
    ObjectImportStatement,
    ModuleImportStatement,
    AnyImportStatement
)

from pyfileconf.io.file.load.parsers.funcdef import extract_function_definition_or_class_init_from_ast_by_name
from pyfileconf.io.file.load.lazy.base.impassign import ImportAssignmentLazyLoader
from pyfileconf.imports.models.statements.container import ImportStatementContainer

ArgumentsAndImports = Tuple[ast.arguments, ImportStatementContainer]

class FunctionArgsExtractor:

    def __init__(self, object_view: 'ObjectView'):
        self.object_view = object_view

    def extract_args(self) -> ArgumentsAndImports:
        return extract_function_args_and_arg_imports_from_import(
            self.object_view.name,
            self.object_view.import_statement,
            self.object_view.section_path_str
        )


def extract_function_args_and_arg_imports_from_import(function_name: str, imp: Optional[AnyImportStatement],
                                                      import_section_path_str: Optional[str] = None) -> ArgumentsAndImports:
    from pyfileconf.io.func.load.extractimp import extract_import_statements_from_function_args_imports_and_assigns

    if imp is None:
        raise NoImportStatementException('no import passed to extract function and imports')

    filepath = get_module_filepath_from_import(
        imp,
        import_section_path_str
    )

    if function_name in imp.renames.new_names:
        # If was renamed in the import,
        # Get original name from rename statement
        function_name = imp.renames.reverse_name_map[function_name]

    loader = ImportAssignmentLazyLoader(filepath)

    # TODO [#44]: handle relative nested imports such as from .this import that
    #
    # This code keeps importing and checking whether the name is defined in the file. If not,
    # it follows the next import. This doesn't properly work if one of those followed imports
    # is a relative import such as `from .this import that`. Add test to show this and
    # also fix it.

    ast_function_def = extract_function_definition_or_class_init_from_ast_by_name(
        loader.ast,
        function_name
    )

    if ast_function_def is not None:
        # Found function definition in this import
        # TODO [#12]: handle other class methods - also update extract_function_definition function
        if ast_function_def.name == '__init__':
            # Got class init
            del ast_function_def.args.args[0]  # delete first arg (self)
        function_arg_imports = extract_import_statements_from_function_args_imports_and_assigns(
            ast_function_def.args,
            loader.imports,
            loader.assigns,
            imp.module
        )
        return ast_function_def.args, function_arg_imports

    # Else, this function must have been imported into this file as well.
    # Must find the matching import, the extract args from that import
    next_level_import = loader.imports.get_import_for_module_or_obj_name(function_name)
    if next_level_import is None:
        raise ValueError(f'not able to find function definition for {function_name}. Could not '
                         f'find in definitions or in imports')

    if import_section_path_str is not None:
        next_module = next_level_import.module
        new_section_path = resolve_name(next_module, import_section_path_str) if next_module.startswith('.') else next_module
    else:
        new_section_path = None

    return extract_function_args_and_arg_imports_from_import(
        function_name,
        imp=next_level_import,
        import_section_path_str=new_section_path
    )


def get_module_filepath_from_import(imp: Optional[AnyImportStatement], import_section_path_str: Optional[str]):
    if imp is None:
        return None

    # TODO [#30]: figure out why mypy won't pick up that imp is not None in get_module_filepath_from_import
    #
    # Then type ignores can be removed
    if isinstance(imp, ObjectImportStatement):
        func = lambda: imp.get_module_filepath(import_section_path_str)
    elif isinstance(imp, ModuleImportStatement):
        # have already ensured in creating ObjectView that there should be only one module
        # in the module import statement. So just take the first filepath
        func = lambda: imp.get_module_filepaths(import_section_path_str)[imp.modules[0]]  # type: ignore
    else:
        raise ValueError(f'expected import statement to be ObjectImportStatement or ModuleImportStatement.'
                         f' Got {imp} of type {type(imp)}')

    return func()


import pluggy


def get_plugin_manager() -> pluggy.PluginManager:
    from pyfileconf.plugin import hookspecs, default_hooks

    plm = pluggy.PluginManager("pyfileconf")
    plm.add_hookspecs(hookspecs)
    plm.load_setuptools_entrypoints("pyfileconf")
    plm.register(default_hooks)
    return plm


def reset_plugins():
    from pyfileconf.plugin import manager

    plm = get_plugin_manager()
    manager.plm = plm


def remove_default_plugins():
    from pyfileconf.plugin import manager

    manager.plm.unregister(name='pyfileconf.plugin.default_hooks')


import ast
from typing import Union, TYPE_CHECKING
if TYPE_CHECKING:
    from pyfileconf.assignments.models.statement import AssignmentStatement
    from pyfileconf.assignments.models.container import AssignmentStatementContainer

AssignmentStatementOrNone = Union['AssignmentStatement', None]

class AssignmentExtractor(ast.NodeVisitor):

    def __init__(self):
        from pyfileconf.assignments.models.container import AssignmentStatementContainer
        self.assigns = AssignmentStatementContainer([])

    def visit_Assign(self, node):
        from pyfileconf.assignments.models.statement import AssignmentStatement
        self.assigns.append(
            AssignmentStatement.from_ast_assign(node)
        )

    def visit_AnnAssign(self, node):
        from pyfileconf.assignments.models.statement import AssignmentStatement
        self.assigns.append(
            AssignmentStatement.from_ast_assign(node)
        )


def extract_assignments_from_ast(module: ast.Module) -> 'AssignmentStatementContainer':
    ae = AssignmentExtractor()
    ae.visit(module)
    return ae.assigns

def extract_assignment_from_ast(module: ast.Module) -> AssignmentStatementOrNone:
    assign_container = extract_assignments_from_ast(module)

    if len(assign_container) == 0:
        return None

    if len(assign_container) > 1:
        raise ValueError(f'expected to extract one assignment from ast. got {len(assign_container)} '
                         f'assigns: {assign_container.items}')

    return assign_container[0]

class AssignmentByVarnameExtractor(AssignmentExtractor):

    def __init__(self, varname: str):
        self.varname = varname
        super().__init__()

    def visit_Assign(self, node):
        varnames = [target.id for target in node.targets]
        if self.varname in varnames:
            super().visit_Assign(node)

    def visit_AnnAssign(self, node):
        if node.target.id == self.varname:
            super().visit_AnnAssign(node)


def extract_assignments_from_ast_by_name(module: ast.Module, name: str) -> 'AssignmentStatementContainer':
    ae = AssignmentByVarnameExtractor(name)
    ae.visit(module)
    return ae.assigns

def extract_assignment_from_ast_by_name(module: ast.Module, name: str) -> AssignmentStatementOrNone:
    assign_container = extract_assignments_from_ast_by_name(module, name)

    if len(assign_container) == 0:
        return None

    if len(assign_container) > 1:
        raise ValueError(f'expected to extract one assignment from ast. got {len(assign_container)} '
                         f'assigns: {assign_container.items}')

    return assign_container[0]


from typing import List, Any, cast

from pyfileconf.imports.logic.load.name import get_module_and_name_imported_from, get_imported_obj_variable_name
from pyfileconf.logic.inspect import _is_builtin


def _key_value_pair_to_assignment_str(key: str, value: Any, module_strs: List[str] = None):
    value = _assignment_output_repr(value, module_strs=module_strs)
    return f'{key} = {value}'


def _assignment_output_repr(value: Any, module_strs: List[str] = None):
    """
    Main formatting function to produce executable python code
    Args:
        value: value to be assigned to a variable

    Returns:

    """
    from pyfileconf.selector.models.selector import _is_selector
    from pyfileconf.selector.models.itemview import _is_item_view, ItemView
    if _is_selector(value):
        # accessing porperties of selector object will cause issues. Only need to return Selector()
        return 'Selector()'

    if _is_item_view(value):
        # accessing properties of ItemView object will also cause issues. need to return s. and section path
        return f's.{value.section_path_str}'

    # Handle functions, other things with builtin names
    try:
        return value.__name__
    except (AttributeError, KeyError):
        pass

    # Handle builtins
    if _is_builtin(value):
        return _assignment_output_repr_for_builtins(value)

    # All others, try to get variable name and import
    if module_strs is not None:
        module_and_module_name_or_none = get_module_and_name_imported_from(value, module_strs)
        if module_and_module_name_or_none is None:
            # warnings.warn(f'could not find import module for {value}\n will generate assignment statement. '
            #               f'likely, the configuration file will need to be manually updated.')
            return None

        module, module_name = module_and_module_name_or_none
        variable_name = get_imported_obj_variable_name(value, module)
    else:
        variable_name = value

    # warnings.warn(f'could not find __name__ of type, and type was not builtin for {value} of type {type(value)}.'
    #               f'guessed {variable_name} as value, but may not execute correctly')

    return variable_name


def _assignment_output_repr_for_builtins(value: Any) -> Any:
    if isinstance(value, str):
        return f"r'{value}'"

    # Other builtins, output as is
    return value


from typing import Tuple, List
import ast

from pyfileconf.io.file.load.parsers.base import FileParser

AstModuleAndStrListTuple = Tuple[ast.Module, List[str]]

class PythonFileParser(FileParser):

    def load(self) -> AstModuleAndStrListTuple:  # type: ignore
        # Get text of file
        file_lines = super().load()

        # Parse into ast
        module: ast.Module = ast.parse(''.join(file_lines))

        return module, file_lines


from pyfileconf.io.file.load.lazy.base.loader import LazyLoader
from pyfileconf.io.file.load.parsers.imp import extract_imports_from_ast
from pyfileconf.io.file.load.parsers.assign import extract_assignments_from_ast
from pyfileconf.assignments.models.container import AssignmentStatementContainer
from pyfileconf.imports.models.statements.container import ImportStatementContainer



class ImportAssignmentLazyLoader(LazyLoader):

    def register(self):
        # Store ast representation of file and file body
        super().register()

        # Store imports and assignments
        if self._ast is not None:
            self._imports = extract_imports_from_ast(self._ast)
            self._assigns = extract_assignments_from_ast(self._ast)
        else:
            self._imports = ImportStatementContainer([])
            self._assigns = AssignmentStatementContainer([])

    @property
    def imports(self) -> ImportStatementContainer:
        return self._try_getattr_else_register('_imports')

    @property
    def assigns(self) -> AssignmentStatementContainer:
        return self._try_getattr_else_register('_assigns')


from typing import List, cast
import ast
import warnings

from pyfileconf.io.func.load.config import function_args_as_arg_and_annotation_dict
from pyfileconf.io.file.load.parsers.extname import extract_unique_external_names_from_assign_value
from pyfileconf.imports.models.statements.container import ImportStatementContainer, ObjectImportStatement
from pyfileconf.assignments.models.container import AssignmentStatementContainer
from pyfileconf.logic.inspect import _is_str_matching_builtin_type

def extract_import_statements_from_function_args_imports_and_assigns(args: ast.arguments,
                                                                     imports: ImportStatementContainer,
                                                                     assigns: AssignmentStatementContainer,
                                                                     current_module_section_path_str: str,
                                                                     ) -> ImportStatementContainer:
    defaults_dict, annotation_dict = function_args_as_arg_and_annotation_dict(args)

    # Extract external names. These are the root names (not attrs, the base of attrs) for anything that is not builtin
    external_names = _unique_external_names_from_default_dict_and_annotation_dict(defaults_dict, annotation_dict)

    out_imports = ImportStatementContainer([])

    # External names may be either due to assignment or due to imports
    # Handle names due to assignment
    for name in external_names:
        if assigns.contains_varname(name):
            # External name is due to assignment in this module. Create import from this module
            out_imports.append(ObjectImportStatement.from_str(f'from {current_module_section_path_str} import {name}'))

    # Handle names which are imported into this file
    for name in external_names:
        import_or_none = imports.get_import_for_module_or_obj_name(name)
        if import_or_none is not None:
            out_imports.append(import_or_none)

    # Sanity check, did we find all the imports?
    n_external = len(external_names)
    n_imports = len(out_imports)
    if n_external != n_imports:
        warnings.warn(f'had {n_external} external names from function definition, only '
                      f'found {n_imports} imports. May be missing imports')

    return out_imports

def _unique_external_names_from_default_dict_and_annotation_dict(default_dict: dict,
                                                                 annotation_dict: dict) -> List[str]:
    names = _unique_external_names_from_assignment_dict(default_dict)
    names += _unique_external_names_from_annotation_dict(annotation_dict)

    return list(set(names))

def _unique_external_names_from_multiple_assignment_dicts(*assignment_dicts) -> List[str]:
    all_names: List[str] = []
    for assignment_dict in assignment_dicts:
        all_names += _unique_external_names_from_assignment_dict(assignment_dict)
    return list(set(all_names))

def _unique_external_names_from_assignment_dict(assignment_dict: dict) -> List[str]:
    names = []
    for key, value in assignment_dict.items():
        new_names = extract_unique_external_names_from_assign_value(value)
        names.extend(new_names)
    no_none_names: List[str] = [name for name in names if name is not None]
    return list(set(no_none_names))

def _unique_external_names_from_annotation_dict(annotation_dict: dict) -> List[str]:
    type_str_names = _extract_unique_type_str_names_from_annotation_dict(annotation_dict)
    return [type_str for type_str in type_str_names if not _is_str_matching_builtin_type(type_str)]

def _extract_unique_type_str_names_from_annotation_dict(annotation_dict: dict) -> List[str]:
    names = []
    for value in annotation_dict.values():
        names.extend(_extract_str_names_from_ambiguous_annotation(value))

    return list(set(names))

def _extract_str_names_from_ambiguous_annotation(annotation) -> List[str]:
    # TODO [#26]: remove type ignores once typeshed has better ast support
    #
    # Hitting errors "expr" has no attribute "id" and  "slice" has no attribute "value"

    names = []
    if isinstance(annotation, ast.Name):
        names.append(annotation.id)  # handles most types
    elif hasattr(annotation, 'elts'):  # type: ignore
        # got multiple values, e.g. List[str, bool]
        # in this case, value.slice.value is a Tuple, and Tuple.elts contains the items
        names.extend(_extract_str_names_from_tuple(annotation)) # type: ignore
    elif isinstance(annotation, ast.Index):
        # e.g. [str]
        # use .value to extract the str part, but then as it can be anything, call recursively on this
        names.extend(_extract_str_names_from_ambiguous_annotation(annotation.value))
    elif isinstance(annotation, ast.Subscript):
        # E.g. Tuple[int]
        names.extend(_extract_str_names_from_subscript(annotation))
    elif isinstance(annotation, ast.Attribute):
        # E.g. pd.DataFrame
        names.append(annotation.value.id)  # type: ignore
    elif isinstance(annotation, ast.Str):
        # E.g. 'DataPipeline' (quoted type declaration where type was imported if TYPE_CHECKING)
        names.append(annotation.s)
    else:
        raise NotImplementedError(f'no handling for {annotation} of type {type(annotation)}')

    return names

def _extract_str_names_from_subscript(subscript: ast.Subscript) -> List[str]:
    names = []

    # for example, List[str]
    outer_ast = subscript.value  # List portion
    inner_ast = subscript.slice  # [str] portion

    names.extend(_extract_str_names_from_ambiguous_annotation(outer_ast))
    names.extend(_extract_str_names_from_ambiguous_annotation(inner_ast))

    return names

def _extract_str_names_from_tuple(tuple: ast.Tuple) -> List[str]:
    names = []
    for item in tuple.elts:
        names.extend(_extract_str_names_from_ambiguous_annotation(item))
    return names


from collections import defaultdict
from typing import Dict, Set, Optional, TYPE_CHECKING, Union


if TYPE_CHECKING:
    from pyfileconf.main import PipelineManager
    from pyfileconf.selector.models.itemview import ItemView
    from pyfileconf.sectionpath.sectionpath import SectionPath


class PyFileConfContext:
    config_dependencies: Dict[str, Set['SectionPath']]
    active_managers: Dict[str, 'PipelineManager']
    force_update_dependencies: Dict[str, Set['SectionPath']]
    file_is_currently_being_loaded: bool
    currently_running_section_path_str: Optional[str]

    def __init__(self, config_dependencies: Optional[Dict[str, Set['SectionPath']]] = None,
                 active_managers: Optional[Dict[str, 'PipelineManager']] = None,
                 force_update_dependencies: Optional[Dict[str, Set['SectionPath']]] = None,
                 file_is_currently_being_loaded: bool = False,
                 currently_running_section_path_str: Optional[str] = None):
        if config_dependencies is None:
            config_dependencies = defaultdict(lambda: set())
        if active_managers is None:
            active_managers = {}
        if force_update_dependencies is None:
            force_update_dependencies = defaultdict(lambda: set())

        self.config_dependencies = config_dependencies
        self.active_managers = active_managers
        self.force_update_dependencies = force_update_dependencies
        self.file_is_currently_being_loaded = file_is_currently_being_loaded
        self.currently_running_section_path_str = currently_running_section_path_str

    def reset(self):
        self.config_dependencies = defaultdict(lambda: set())
        self.active_managers = {}
        self.force_update_dependencies = defaultdict(lambda: set())
        self.file_is_currently_being_loaded = False
        self.currently_running_section_path_str = None

    def add_config_dependency(self, dependent: Union[str, 'ItemView', 'SectionPath'],
                              depends_on: Union[str, 'ItemView', 'SectionPath'], force_update: bool = False):
        from pyfileconf.sectionpath.sectionpath import SectionPath
        dependent_section_path = SectionPath.from_ambiguous(dependent)
        depends_on_section_path_str = SectionPath.from_ambiguous(depends_on).path_str
        self.config_dependencies[depends_on_section_path_str].add(dependent_section_path)
        if force_update:
            self.force_update_dependencies[depends_on_section_path_str].add(dependent_section_path)

    def add_config_dependency_for_currently_running_item_if_exists(self, depends_on: Union[str, 'ItemView', 'SectionPath'],
                                                                   force_update: bool = False):
        if self.currently_running_section_path_str is not None:
            self.add_config_dependency(
                self.currently_running_section_path_str, depends_on, force_update=force_update
            )

    def add_config_dependency_if_file_is_currently_being_loaded(
        self,
        dependent: Union[str, 'ItemView', 'SectionPath'],
        depends_on: Union[str, 'ItemView', 'SectionPath'],
        force_update: bool = False
    ):
        if self.file_is_currently_being_loaded:
            self.add_config_dependency(dependent, depends_on, force_update=force_update)


context = PyFileConfContext()



import importlib.util
from importlib.abc import Loader
from types import ModuleType
import sys
from typing import Optional

from pyfileconf.exceptions.imports import CouldNotImportException
from pyfileconf.imports.models.statements.container import ImportStatementContainer

def get_user_defined_dict_from_filepath(filepath: str, module_name: str=None, remove_imports=False,
                                        imports: ImportStatementContainer=None) -> dict:
    module = _load_file_as_module(filepath, name=module_name)
    user_defined_dict = _get_user_defined_dict_from_module(module)
    if remove_imports:
        if imports is None:
            raise ValueError('must pass imports when passing remove_imports=True')
        return {
            key: value for key, value in user_defined_dict.items() \
            if imports.get_import_for_module_or_obj_name(key) is None
        }
    else:
        return user_defined_dict

def _get_user_defined_dict_from_module(module: ModuleType) -> dict:
    out_dict = {}
    for key, value in module.__dict__.items():
        if key.startswith('__') or isinstance(value, ModuleType):
            continue
        out_dict.update({key: value})

    return out_dict


def _load_file_as_module(filepath: str, name: Optional[str] = 'module.name') -> ModuleType:
    if name is None:
        name = 'module.name'
        add_to_sys = False
    else:
        add_to_sys = True

    spec = importlib.util.spec_from_file_location(name, filepath)
    module = importlib.util.module_from_spec(spec)
    if spec is None or spec.loader is None or not isinstance(spec.loader, Loader):
        raise CouldNotImportException(
            f'got no spec.loader after calling importlib.util.spec_from_file_location on {filepath}'
        )
    spec.loader.exec_module(module)

    if add_to_sys:
        sys.modules.update({name: module})

    return module


from typing import Tuple, Type, List, Dict, Union

from pyfileconf.main import PipelineManager, SpecificClassConfigDict


def get_pipeline_dict_path_and_specific_class_config_dicts_from_manager(manager: PipelineManager
                                                                        ) -> Tuple[
                                                                                str,
                                                                                List[SpecificClassConfigDict]
                                                                        ]:
    return manager.pipeline_dict_path, manager.specific_class_config_dicts


from typing import Type, Any, Dict

from pyfileconf.imports.logic.load.func import function_args_as_dict


def class_function_args_as_dict(klass: Type) -> Dict[str, Any]:
    """
    To be used on regular methods of a class e.g. ExampleClass.__init__, not on methods of an instance of the class
    """
    arg_dict = function_args_as_dict(klass.__init__)

    # TODO [#35]: handle removing first argument from __init__, may not be named self
    #
    # Current implementation depends on the argument being named self. This seems to
    # be a hard problem in general without a great solution. See
    # https://stackoverflow.com/a/47599893

    # Remove self from arguments
    arg_dict = {key: value for key, value in arg_dict.items() if key != 'self'}

    return arg_dict
